/*
   Copyright 2009 Andrea Pagnani and Alfredo Braunstein

   This file is part of DILPERC (DILuted discrete PERCeptron analysis tool).

   DILPERC is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License Version 2 
   as published by the Free Software Foundation 

   DILPERC is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with DILPERC; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#include <mex.h>

#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "bp.h"


extern bool utIsInterruptPending();

int converge(int T)
{
	int t;
	double diff = 0;
	for (t = 0; t < T; ++t) {
		diff = iterate();
		mexPrintf("err: %i %g\n", t, diff);
		if (diff < err)
			return 1;
		if (utIsInterruptPending()){
		        mexPrintf("err: %i %g\n", t, diff);
			free_memory();
			mexErrMsgTxt("Caught a CTRL-C exiting ...\n");
		}
	}
	mexPrintf("err: %i %g\n", t, diff);
	return 0;
}

int paracool(int T, int Q, double ** myP, double *oldP, double *newP,  double *entr)
{
	int t,n, q,  flagconverge;
	double diff = 0;
	double oldS, newS;

	oldP = calloc(2*N*(Q+1), sizeof(double));
	newP = calloc(2*N*(Q+1), sizeof(double));

	flagconverge=converge(T);
	newS = entropy();
        for (t = 0; t < T; ++t){
		oldS = newS;
		for (q = 0; q < 2 * Q + 1 ; ++q)
			for( n = 0; n<N; ++n )
				oldP[q*N + n] =  P[n][q];

		flagconverge=converge(T);
		newS = entropy();
		for (q = 0; q < 2 * Q + 1 ; ++q)
                        for( n = 0; n<N; ++n )
                                newP[q*N + n] =  P[n][q];


		hdil += deltahdil;
		for(n=0; n<N; ++n )
			Hext[n] = hdil;

		if (utIsInterruptPending()){
			free_memory();
			mexErrMsgTxt("Caught a CTRL-C exiting ...\n");
		}
		mexPrintf("paracool: %i %g %g %g\n", t, newS, hdil, diff);
		mexEvalString("drawnow"); 
		if (newS < 0.05 && newS > 0 && flagconverge ){
			*entr = newS;
			*myP = newP;
                        break;
		}
		if (newS < 0 || flagconverge == 0 ){
			*entr = oldS;
			*myP = oldP;
			break;
		}

        }
	mexPrintf("paracool-final: %i %g %g %g\n", t, *entr, hdil, diff);
	mexEvalString("drawnow"); 
	return 0;
}


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	int n, m, q, maxiter, linearstringlength, paracoolstringlength, paracooling = 0;
	mxArray * xData;
	mxArray * qData;
	mxArray * hData;
	mxArray * tData;
	mxArray * epsData;
	mxArray * linearData;
	mxArray * paracoolData;
	mxArray * deltahData;
	char *linearstring = NULL;
	char *paracoolstring = NULL;
	double * xValues, *outP, *outS;
	double *oldP = NULL, *newP = NULL, *myP=NULL;
	double S;

	/*	printf("nrhs = %d\n", nrhs); */

	if( (nrhs!=6) && (nrhs != 8) ) {
		mexErrMsgTxt("Six inputs required for simple convergence: data-matrix Q H-dil maxiter convergence \"linear\" \nor\nEight inputs required for paracooling: data-matrix Q H-dil maxiter convergence \"linear\"  \"paracooling\" h0 deltah  ");
	}




	xData = (mxArray *) prhs[0];
	xValues = mxGetPr(xData);

	qData=(mxArray *) prhs[1];
	Q = (int) (mxGetScalar(qData));


	hData=(mxArray *) prhs[2];
	hdil = (double) (mxGetScalar(hData));

	tData=(mxArray *) prhs[3];
	maxiter=(int) (mxGetScalar(tData));

	epsData=(mxArray *) prhs[4];
	err = (double) (mxGetScalar(epsData));

	linearData=(mxArray *) prhs[5];
	linearstringlength = mxGetN(linearData)+1;
	linearstring=mxCalloc( linearstringlength, sizeof(char));
	mxGetString(linearData, linearstring, linearstringlength);
	if ( strcmp(linearstring, "linear") )
		nonlinear = 1;
	else
		nonlinear = 0;

	err = (double) (mxGetScalar(epsData));

	if ( nrhs == 8 ){
		paracoolData=(mxArray *) prhs[6];
		deltahData = (mxArray *) prhs[7];
		paracoolstringlength = mxGetN(paracoolData) + 1;
		paracoolstring = mxCalloc(paracoolstringlength, sizeof(char));
		mxGetString(paracoolData, paracoolstring, paracoolstringlength);

		if ( ! strcmp(paracoolstring, "paracooling" ) )
			paracooling =  1;
		else
			mexErrMsgTxt("only paracooling method implemented");
		deltahdil = (double ) mxGetScalar(deltahData);
	}


	N = mxGetN(xData) - 1 ;
	M = mxGetM(xData);

	usrand(123);
	init();

	for (m = 0; m < M; ++m){ 
		Y[m] = xValues[m]; 
		/* printf("Y[%i] = %g\n", m,Y[m]); */
	}

	for (n = 1; n <= N; ++n) {
		for (m = 0; m < M; ++m) {
			X[m][n-1] = xValues[n*M+m];
			/*printf("%e\n", xValues[n*M+m]);*/
		}
	}

	plhs[0] = mxCreateDoubleMatrix(N, 2*Q+1, mxREAL);
	plhs[1] = mxCreateDoubleMatrix(1,1,mxREAL);


	if( paracooling ){
		paracool(maxiter,Q,  &myP, oldP, newP, &S);
		printf("ncore = %i\n", omp_get_max_threads());
		outP = mxGetPr(plhs[0]);
		outS = mxGetPr(plhs[1]);
		for (q = 0; q < 2 * Q + 1 ; ++q)
			for( n = 0; n<N; ++n )
				outP[q*N + n] =  myP[q*N + n];

		outS[0] = S;
		free(oldP);
		free(newP);
	}
	else{
		converge(maxiter);
		printf("ncore = %i\n", omp_get_max_threads());
		printf("N: %i, M: %i Q: %i S: %g\n", N, M,Q, (S=entropy()));
		outP = mxGetPr(plhs[0]);
		outS = mxGetPr(plhs[1]);
		for (q = 0; q < 2 * Q + 1 ; ++q)
			for( n = 0; n<N; ++n ) 
				outP[q*N + n] =  P[n][q];

		outS[0] = S;
	}
	free_memory();
	return;
}


